import java.util.*;
import java.io.*;

public class BeliefNet implements Serializable {

    final String _TRUE = "=true";
    final String _FALSE = "=false";
    
    protected int _nn;               //the number of Nodes in this Belief Net
    protected Node[] _ns;            //stores the Node objects of this Belief Net
    protected Hashtable globalCPT;  //stores the local CPTs of all the Nodes
                                     //also stores temporary factors that are calculated
                                     //during exact inference

    public BeliefNet(Node[] ns) {
	_nn = ns.length;
	_ns = ns;
	globalCPT = new Hashtable();

	for (int i=0; i<_nn; i++) {
	    globalCPT.putAll(_ns[i].getCPT());
	}//for
    }//constructor


    //Constructor for Class BeliefNet: instantiates new Node objects
    public BeliefNet() {
	try {
	    //stream to read from stdin
	    InputStreamReader isr = new InputStreamReader(System.in);
	    BufferedReader bf = new BufferedReader(isr);

	    System.out.print("How many variables (or nodes) does this BayesNet have ?");
	    _nn = Integer.parseInt(bf.readLine());
	    _ns = new Node[_nn];
	    globalCPT = new Hashtable();
            
	    for (int i=0; i<_nn; i++) {
		System.out.println("\nNode "+(i+1));
		System.out.print("What is the name of the new variable(or Node)? ");
		_ns[i] = new Node(bf.readLine());
		globalCPT.putAll(_ns[i].getCPT());
	    }//for

	}//try
	catch (IOException ioe) {
	    ioe.printStackTrace();
	}//catch

    }//Constructor

    public int findNode(String nodeName) {
	for (int i =0; i<_nn; i++)
	    if (_ns[i].getNodeName().equals(nodeName)) 
		return i;
	System.err.println("BeliefNet.findNode: node not found!");
	System.exit(-1);
	return -1;
    }

    public String[] getParents(String nodeName) {
	int nodeIndex = findNode(nodeName);
	return _ns[nodeIndex].getParents();
    }

    public String[] getChildren(String nodeName) {
	int nodeIndex = findNode(nodeName);
	return _ns[nodeIndex].getChildren();
    }

    public Hashtable getCPT(String nodeName) {
	int nodeIndex = findNode(nodeName);
	return _ns[nodeIndex].getCPT();
    }

    //As an example: the string representation of rain would be rain|cloudy
    public String getStringRepresentation(String nodeName) {
	int nodeIndex = findNode(nodeName);
	String[] parents = _ns[nodeIndex].getParents();
	if (parents.length == 0) return nodeName;
	String r = nodeName+"|";

	for (int i=0; i<parents.length; i++) {
	    r = r.concat(parents[i]);
	    if (i+1 < parents.length)
		r = r.concat("&");
	}//for
	return r;
    }

    //a vector representation is the string representation of all the nodes
    public Vector getVectorRepresentation() {
	Vector r = new Vector(_nn);

	for (int i=0; i<_nn; i++)
	    r.add(getStringRepresentation(_ns[i].getNodeName()));
	return r;
    }//getVectorRepresentation

    //this is necessary to get rid of all the factors that have been added
    //using addFactorToGlobalCPT and are therefore specific to a certain
    //inference procedure call.
    public void reinitialiseGlobalCPT() {
	globalCPT.clear();
	for (int i=0; i<_nn; i++)
	    globalCPT.putAll(_ns[i].getCPT());
    }//reinitialiseGlobalCPT

    public void addFactorToGlobalCPT(String key, String prob) {
	globalCPT.put((Object)key,(Object)prob);
    }//addFactorToGlobalCPT

    public double findProbInGlobalCPT(String instantiatedRep) 
    {
        String[] allpermutations = Permutations(instantiatedRep);
        //for (int i=0; i<allpermutations.length; i++)
        //    System.out.println((i+1)+" "+allpermutations[i]);
        
        String prob = "";
        for (int i=0; i<allpermutations.length; i++) {
            prob = (String)globalCPT.get(allpermutations[i]);
            if (prob != null) break;
        }
	if (prob==null) {
            return 1.0;
	}
	double r = Double.parseDouble(prob);
	return r;
    }//findProbInGlobalCPT

    //the query can be instantiated or not and is of the form
    //rain|cloudy&wetgrass or just cloudy&wetgrass etc...
    public String[] Permutations(String strRep) 
    {
        StringTokenizer st = new StringTokenizer(strRep,"|");
        int numVars = st.countTokens();

        if (numVars==1) return andedPermutations(strRep);

        String query = st.nextToken();
        String evidence = st.nextToken();

        String[] qpermute = andedPermutations(query);
        String[] epermute = andedPermutations(evidence);
        
        int numtotalpermutations = qpermute.length*epermute.length;
        String[] allpermutations = new String[numtotalpermutations];
        int counter=0;
        for (int i=0; i<qpermute.length; i++) {
            for (int j=0; j<epermute.length; j++) {
                allpermutations[counter]=qpermute[i]+"|"+epermute[j];
                counter++;
            }//for
        }//for
        return allpermutations;
    }//Permutations
    
    //can be passed instantiated or unintantiated strings
    //that have only been anded, ie rain&cloudy
    public String[] andedPermutations(String andedStr) {
	StringTokenizer st = new StringTokenizer(andedStr, "&");
	int numVars = st.countTokens();

	String[] strArray = new String[numVars];
	strArray[0] = andedStr;

	for (int i=1; i<numVars; i++) {
	    st = new StringTokenizer(strArray[i-1],"&");
	    String varInFront = st.nextToken();
	    String minusVarInFront = strArray[i-1].substring(varInFront.length()+1,strArray[i-1].length());
	    strArray[i] =  minusVarInFront+"&"+varInFront;
	}
	return strArray;
    }//returnPermutations

    //the vector v MUST contain instantiated factors
    public double multiplyVector(Vector v) {
	double r = 1.0;
	Enumeration _e = v.elements();
	while (_e.hasMoreElements()) {
	    String factor = (String)_e.nextElement();
	    r *= findProbInGlobalCPT(factor);
	}//while
	return r;
    }//multiplyVector
    
    /**********************************************************/
    /****  Method int2binary - converts an integer to its  ****/
    /****  binary value and appends zeros to make it as    ****/
    /****  long as the parameter length. The binary number ****/
    /****  is stored in an integer array, index 0 is the   ****/
    /****  most significant digit. Zeros are appended      ****/
    /****  before the most significant digit               ****/
    /**********************************************************/
    public int[] int2binary (int value,int length) {
        String binary ="";
        int[] rem = new int[length];
	for (int h=0; h<length; h++) rem[h] = 0;

	for (int i=length-1; value > 0; i--) {

	    rem[i] = value%2;
	    value = value/2;
	}        
	return rem;
    }

    //this method is used in class Menu
    public Node[] getNodeStructure() {
        return _ns;
    }



    /*********************  UNUSED METHODS  ***********************/



    /**************************************************************/
    /*** Generates all possible instantiations for strRep       ***/
    /*** and returns a hashtable of probabilities hashed to     ***/
    /*** a particular instantiation. This method should not be  ***/
    /*** called with an instantiated representation, use the    ***/
    /*** method findProbInGlobalCPT instead.                    ***/
    /**************************************************************/
    public Hashtable lookUpGlobalCPT(String strRep) {
        Hashtable ht = new Hashtable();
        StringTokenizer st = new StringTokenizer(strRep,"|");

        String[] qinst; //holds all possible instantiations of query vars
        String[] einst; //holds all possible instantitations of evidence vars
        String[] qvars; //holds all query vars
        String[] evars; //holds all evidence vars
        
        
        StringTokenizer qst = new StringTokenizer(st.nextToken(),"&");
        int varsQuery = qst.countTokens();
        int instQuery = (int)Math.pow(2,varsQuery);
        qinst = new String[instQuery];
        qvars = new String[varsQuery];
        
        for(int i=0; i<varsQuery; i++)
            qvars[i] = qst.nextToken();
        
        for (int i=0; i<instQuery; i++) {
            int[] binary = int2binary(i,varsQuery);
            qinst[i] = "";
            for (int j=0; j<varsQuery; j++) {
                if (binary[j]==0) qinst[i] = qinst[i].concat(qvars[j]+_FALSE);
                else qinst[i] = qinst[i].concat(qvars[j]+_TRUE);
                
                if (j+1 < varsQuery) qinst[i] = qinst[i].concat("&");
            }//inner for
        }//outer for

        if (st.hasMoreTokens()) {
            StringTokenizer est = new StringTokenizer(st.nextToken(),"&");
            int varsEvidence = qst.countTokens();
            int instEvidence = (int)Math.pow(2,varsQuery);
            einst = new String[instEvidence];
            evars = new String[varsEvidence];
        
            for(int i=0; i<varsEvidence; i++)
                evars[i] = est.nextToken();
            
            for (int i=0; i<instEvidence; i++) {
                int[] binary = int2binary(i,varsEvidence);
                einst[i] = "";
                for (int j=0; j<varsEvidence; j++) {
                    if (binary[j]==0) einst[i] = einst[i].concat(evars[j]+_FALSE);
                    else einst[i] = einst[i].concat(evars[j]+_TRUE);
                    
                    if (j+1 < varsEvidence) einst[i] = einst[i].concat("&");
                }//inner for
            }//outer for
            
            //make all combinations of query and evidence variables
            String[] allcomb = new String[qinst.length*einst.length];
            int counter = 0;
            for (int i=0; i<qinst.length; i++) {
                for (int j=0; j<einst.length; j++) {
                    allcomb[counter] = qinst[i]+"|"+einst[j];
                    counter++;
                }//innner for
            }//outer for

            //lookup probabilities for all entries in allcomb and make a hashtable
            Hashtable rht = new Hashtable(allcomb.length);
            for (int i=0; i<allcomb.length; i++)
                rht.put((Object)allcomb[i],(Object)new Double(findProbInGlobalCPT(allcomb[i])));
            return rht;
        }//if

        //there is no evidence so return hashtable of all instantiations of
        //the query variables hashed to their probabilities
        Hashtable rht = new Hashtable(qinst.length);
        
        for (int i=0; i<qinst.length; i++)
            rht.put((Object)qinst[i],(Object)new Double(findProbInGlobalCPT(qinst[i])));
        return rht;
    }//lookUpGlobalCPT
}//class
